#ifndef __MODEL_DEFINITIONS_H__
#define __MODEL_DEFINITIONS_H__

/*
 * MODEL_DEFINITIONS.H
 */
 
#include "definitions.h"
#include "geometry_definitions.h"

/* ---------- constants */

enum
{
	MODEL_GROUP_TAG= 0x6d6f6465,	// 'mode'
	MODEL_VERSION_NUMBER= 4,
	MODEL_FILE_TYPE= 0x4d4f4445,	// 'MODE'

	MAXIMUM_CELLS_PER_MODEL= 128
};

/* ---------- flags */
enum {
	_vertex_deforms_mesh_bit,
	_vertex_is_media_bit,
	_vertex_wobbles_when_media_bit,
	NUMBER_OF_GEOMETRY_VERTEX_BITS,
	
	_vertex_deforms_mesh_flag= FLAG(_vertex_deforms_mesh_bit),
	_vertex_is_media_flag= FLAG(_vertex_is_media_bit),
	_vertex_wobbles_when_media_flag= FLAG(_vertex_wobbles_when_media_bit)
};

enum {
	_model_mesh_cell_replaces_mesh_bit, // mesh cell under this does not render
	_mesh_cell_is_impassable_walking_bit,
	_mesh_cell_is_impassable_flying_bit,
	_mesh_cell_overlapped_by_model_bit,
	NUMBER_OF_MODEL_MESH_CELL_BITS,
	
	_model_mesh_cell_replaces_mesh_flag= FLAG(_model_mesh_cell_replaces_mesh_bit),
	_mesh_cell_is_impassable_walking_flag= FLAG(_mesh_cell_is_impassable_walking_bit),
	_mesh_cell_is_impassable_flying_flag= FLAG(_mesh_cell_is_impassable_flying_bit),
	_mesh_cell_overlapped_by_model_flag= FLAG(_mesh_cell_overlapped_by_model_bit)
};

enum {
	_model_units_do_not_avoid_bit,
	
	_model_units_do_not_avoid_flag= FLAG(_model_units_do_not_avoid_bit)
};

/* ---------- structures */
#define MATERIAL_NOT_RENDERED_IN_PERMUTATION (UNSIGNED_CHAR_MAX)

#define SIZEOF_STRUCT_MODEL_PERMUTATION_DATA 64
struct model_permutation_data {
	short collection_reference_permutation;

	byte permutations[MAXIMUM_MATERIALS_PER_GEOMETRY]; // permutations for the given material

	short unused[13];
	short collection_index; // loaded at run time
	short color_table_index; // at run time...
};

#define SIZEOF_STRUCT_GEOMETRY_INDEX_DATA (4)
struct geometry_vertex_data {
	long flags;
};

#define SIZEOF_STRUCT_MODEL_MESH_DATA (8) 
struct model_mesh_cell_data {
	short x, y;
	word flags;
	word unused;
};

#define SIZEOF_STRUCT_MODEL_DEFINITION 64
struct model_definition
{
	unsigned long flags;
	
	file_tag geometry_tag;
	short geometry_index;

	// counts	
	short model_geometry_vertex_count;
	short model_permutation_count;
	short model_cell_count;

	// array of geometry_vertex flags...
	int model_geometry_vertex_offset, model_geometry_vertex_size;
	struct geometry_vertex_data *vertex_flags;

	// array of model_permutations
	int model_permutations_offset, model_permutations_size;
	struct model_permutation_data *model_permutations;

	// array of model_mesh_cell_data
	int model_cell_offset, model_cell_size;
	struct model_mesh_cell_data *model_cells;

	// conglomerate data pointer..
	long data_offset, data_size;
	void *data; // memory-based
};

/* ---------- prototypes */

#define initialize_model_definitions() \
	initialize_definitions(&models)
#define initialize_model_definitions_for_new_map() \
	initialize_definitions_for_new_map(&models)

void dispose_model_definitions_from_old_map(void);

boolean load_model_data(long tag_index, struct model_definition *definition, boolean readahead);

#define model_definition_tag_to_model_type(subgroup) \
	definition_tag_to_index(&models, (subgroup))
#define get_model_definition_tag_header(index) get_definition_tag_header(&models, (index))

#ifdef DEBUG
#define get_model_definition(index) ((struct model_definition *)get_definition(&models, (index)))
struct model_permutation_data *get_model_permutation_data(struct model_definition *definition, int permutation);
struct model_mesh_cell_data *get_model_mesh_cell(struct model_definition *definition, int index);
struct geometry_vertex_data *get_model_geometry_vertex_data(struct model_definition *definition, int index);
#else
#define get_model_permutation_data(d, p) ((d)->model_permutations+(p))
#define get_model_definition(index) ((struct model_definition *)(models.definitions)+(index))
#define get_model_mesh_cell(d, index) ((d)->model_cells+(index))
#define get_model_geometry_vertex_data(d, index) ((d)->vertex_flags+(index))
#endif

/* ---------- globals */

extern struct definition_globals models;

#endif