#ifndef __SHAPE_FILE_H__
#define __SHAPE_FILE_H__

/*
SHAPE_FILE.H
Wednesday March 13 1996, 9:22 PM
*/

#ifndef __BITMAPS_H__
#include "bitmaps.h"
#endif

#ifndef __TAG_FILES_H__
#include "tag_files.h"
#endif

#ifndef __WORLD_H__
#include "world.h"
#endif

#ifdef __cplusplus
extern "C" {
#endif

/* ---------- constants */

enum
{
	COLLECTION_GROUP_TAG= 0x2e323536, // '.256' (looks like text)
	COLLECTION_VERSION_NUMBER= 4,

	SHAPE_NAME_LENGTH= 63,
	
	EXTRACTOR_DATA_SIZE= 16
};

enum // user data flags
{
	_collection_can_be_used_for_textures_bit,
	_collection_can_be_used_for_monsters_bit,
	_collection_can_be_used_for_projectiles_bit,
	_collection_can_be_used_for_scenery_bit,
	_collection_can_be_used_for_atmospheric_effect_bit,
	_collection_can_be_used_for_interface_bit,
	_collection_can_be_used_for_unknown3_bit,
	_collection_can_be_used_for_unknown4_bit,
	
	_collection_is_color_map_bit,
	
	NUMBER_OF_COLLECTION_USER_DATA_FLAGS, // <=32
	
	_collection_can_be_used_for_textures_flag= FLAG(_collection_can_be_used_for_textures_bit),
	_collection_can_be_used_for_monsters_flag= FLAG(_collection_can_be_used_for_monsters_bit),
	_collection_can_be_used_for_projectiles_flag= FLAG(_collection_can_be_used_for_projectiles_bit),
	_collection_can_be_used_for_scenery_flag= FLAG(_collection_can_be_used_for_scenery_bit),
	_collection_can_be_used_for_atmospheric_effect_flag= FLAG(_collection_can_be_used_for_atmospheric_effect_bit),
	_collection_can_be_used_for_interface_flag= FLAG(_collection_can_be_used_for_interface_bit),
	_collection_can_be_used_for_unknown3_flag= FLAG(_collection_can_be_used_for_unknown3_bit),
	_collection_can_be_used_for_unknown4_flag= FLAG(_collection_can_be_used_for_unknown4_bit),

	_collection_is_color_map_flag= FLAG(_collection_is_color_map_bit)
};

enum // transfer modes
{
	_transfer_mode_none,
	_transfer_mode_transparent_fade_out, // opaque to fully transparent
	_transfer_mode_transparent_fade_in, // fully transparent to opaque
	_transfer_mode_shrink_out,
	_transfer_mode_shrink_in,
	_transfer_mode_transparent_appearing,
	_transfer_mode_transparent_disappearing,
	_transfer_mode_transparent_cycle,
	
	NUMBER_OF_TRANSFER_MODES
};

/* ---------- structures */

enum // collection_header flags
{
	_collection_header_initialized_bit,
	
	NUMBER_OF_COLLECTION_HEADER_FLAGS,
	
	_collection_header_initialized_flag= FLAG(_collection_header_initialized_bit)
};

#define SIZEOF_STRUCT_COLLECTION_HEADER 320
struct collection_header
{
	// offsets are from beginning of this header

	unsigned long flags;

	unsigned long user_data_flags;

	short unused1[28];

	long color_table_count, color_tables_offset, color_tables_size;
	void *color_tables; // memory-based
	
	long hue_change_count, hue_changes_offset, hue_changes_size;
	struct hue_change *hue_changes; // memory-based
	
	long bitmap_count, bitmap_references_offset, bitmap_references_size;
	struct bitmap_reference *bitmap_references; // memory-based
	
	long bitmap_instance_count, bitmap_instances_offset, bitmap_instances_size;
	struct bitmap_instance_data *bitmap_instances; // memory-based

	long sequence_count, sequence_references_offset, sequence_references_size;
	struct sequence_reference *sequence_references; // memory-based

	long shadow_map_count, shadow_maps_offset, shadow_maps_size;
	struct shadow_map_data *shadow_maps; // memory-based

	long blend_table_count, blend_table_offset, blend_table_size;
	void *blend_table;

	long remapping_table_count, remapping_table_offset, remapping_table_size;
	void *remapping_table;

	long shading_table_count;
	void **shading_table_pointers; // memory-based
	
	short unused3[24];

	long data_offset, data_size, data_delta;
	void *data; // memory-based
	
	short unused2[20];
	
	byte extractor_data[EXTRACTOR_DATA_SIZE];
};

/* ---------- hue changes */

#define MAXIMUM_HUE_CHANGES_PER_COLLECTION 8

enum
{
	_hue_change_primary_bit,
	_hue_change_secondary_bit,
	
	NUMBER_OF_HUE_CHANGE_FLAGS,

	_hue_change_primary_flag= FLAG(_hue_change_primary_bit),
	_hue_change_secondary_flag= FLAG(_hue_change_secondary_bit)
};

#define SIZEOF_STRUCT_HUE_CHANGE 128
struct hue_change
{
	char name[SHAPE_NAME_LENGTH+1];

	word hue0, hue1; // hue0>hue1 is valid for hue ranges crossing 360
	byte colors_effected[PIXEL8_MAXIMUM_COLORS/8];

	word minimum_saturation;
	word median_hue;

	word flags;
	
	short unused[3];
	
	byte extractor_data[EXTRACTOR_DATA_SIZE]; // contains default rgb_color replacement
};

/* ---------- sequences */

/*
shadow maps and light maps should be linked to frames
*/

#define SIZEOF_STRUCT_SEQUENCE_REFERENCE 128
struct sequence_reference
{
	char name[SHAPE_NAME_LENGTH+1];
	long offset, size; // offset from collection header
	struct sequence_data *sequence;
	
	short unused[18];
	
	byte extractor_data[EXTRACTOR_DATA_SIZE];
};

enum // sequence flags
{
	_sequence_uses_mirroring_bit, // right-facing views are mirrored from left-facing views
	_sequence_has_frame_data_bit, // sequence has the new-style frame_data structures (always set)
	_sequence_has_no_rotation_bit, // instead of being different rotations of the same shape
		// each view here is a different permutation of a class (e.g., flying pieces of a body)
	
	NUMBER_OF_SEQUENCE_FLAGS, // <=32
	
	_sequence_uses_mirroring_flag= FLAG(_sequence_uses_mirroring_bit),
	_sequence_has_frame_data_flag= FLAG(_sequence_has_frame_data_bit),
	_sequence_has_no_rotation_flag= FLAG(_sequence_has_no_rotation_bit)
};

struct sequence_frame_data
{
	short shadow_map_index;

	short_world_point3d key_point; // computed from 2d keypoints (?)
	
	short unused[19];
	
//	short bitmap_instance_indexes[]; // .number_of_views
};
#define get_sequence_frame_bitmap_instance_indexes(frame) ((short *)((frame)+1))

enum // frame sound indexes
{
	_first_frame_sound,
	_key_frame_sound,
	_last_frame_sound,
	NUMBER_OF_SEQUENCE_SOUNDS
};

#define SIZEOF_STRUCT_SEQUENCE_DATA 64
struct sequence_data
{
	unsigned long flags;
	
	fixed pixels_to_world;
	
	short number_of_views, frames_per_view, ticks_per_frame;
	short key_frame, loop_frame;
	
	short sound_indexes[NUMBER_OF_SEQUENCE_SOUNDS];
	file_tag sound_tags[NUMBER_OF_SEQUENCE_SOUNDS];
	
	short transfer_mode, transfer_period;

	fixed radius, height0, height1; // height0<height1	
	short_world_distance world_radius, world_height0, world_height1; // (pixels_to_world, above)

	short unused[3];
	
	// frames
};
#define get_sequence_data_frames(sequence) ((struct sequence_frame_data *)((sequence)+1))

/* ---------- bitmap instance */

// if a bitmap instance references a bitmap with a 4-bit alpha channel, it is always transparent,
// otherwise it may be marked transparent and will use the _transparency_overlay where the source
// color component is used as the transparency level.

enum // bitmap instance flags
{
	_bitmap_instance_mirror_horizontal_bit,
	_bitmap_instance_mirror_vertical_bit,
	_bitmap_instance_transparent_bit,
	_bitmap_instance_keypoint_obscured_bit,
	_bitmap_instance_texture_bit, // this is a texture, not a 1-bit transparency encoded rectangle
	
	NUMBER_OF_BITMAP_INSTANCE_FLAGS, // <=32

	_bitmap_instance_mirror_horizontal_flag= FLAG(_bitmap_instance_mirror_horizontal_bit),
	_bitmap_instance_mirror_vertical_flag= FLAG(_bitmap_instance_mirror_vertical_bit),
	_bitmap_instance_transparent_flag= FLAG(_bitmap_instance_transparent_bit),
	_bitmap_instance_keypoint_obscured_flag= FLAG(_bitmap_instance_keypoint_obscured_bit),
	_bitmap_instance_texture_flag= FLAG(_bitmap_instance_texture_bit)
};

#define SIZEOF_STRUCT_BITMAP_INSTANCE_DATA 64
struct bitmap_instance_data
{
	unsigned long flags;

	fixed_fraction minimum_light_intensity, maximum_light_intensity; /* in [0,FIXED_ONE) */
	fixed_fraction transparency; // 0 if !_bitmap_instance_transparent_flag
	short unused1[1];
	
	point2d registration_point; // animation centering
	short_world_distance world_top, world_left, world_right, world_bottom; // (registration_point, pixels_to_world)
	
	point2d key_point; // projectile creation, etc.
	
	short bitmap_index;

	short unused2[9];

	byte extractor_data[EXTRACTOR_DATA_SIZE];
};

/* ---------- shadow maps */

enum // shadow map flags
{
	_shadow_is_light_map_bit, // changes application function
	_shadow_mirror_horizontal_bit,
	_shadow_mirror_vertical_bit,
	
	NUMBER_OF_SHADOW_MAP_FLAGS, // <=32

	_shadow_is_light_map_flag= FLAG(_shadow_is_light_map_bit),
	_shadow_mirror_horizontal_flag= FLAG(_shadow_mirror_horizontal_bit),
	_shadow_mirror_vertical_flag= FLAG(_shadow_mirror_vertical_bit)
};

struct fixed_bitmap_coordinate2d
{
	fixed x, y;
};

#define SIZEOF_STRUCT_SHADOW_MAP_DATA 32
struct shadow_map_data
{
	unsigned long flags;
	
	struct fixed_bitmap_coordinate2d registration_point;
	
	short bitmap_index;

	short unused[1];
	
	byte extractor_data[EXTRACTOR_DATA_SIZE];
};

/* ---------- bitmaps */

#define SIZEOF_STRUCT_BITMAP_REFERENCE 128
struct bitmap_reference
{
	char name[SHAPE_NAME_LENGTH+1];
	long offset, size; // offset from collection header
	struct bitmap_data *bitmap; // NULL if not loaded (not straightforwardly calculated from offset)
	
	short width, height;
	
	short shaded_hardware_addresses[MAXIMUM_NUMBER_OF_SHADED_BITMAP_REFERENCES];
	
	// used when this is a submesh bitmap
	short dirty;
	short disk_cache_index;
	void *base_address; 

	// used for other kinds of bitmaps during hardware acceleration
	void *hardware_format;
	unsigned long hardware_address;
	short hardware_texture_cache_index;
	word pad;
	
	byte extractor_data[EXTRACTOR_DATA_SIZE];
};

/* ---------- prototypes/SHAPE_FILE.C */

void assert_shape_file_constants(void);

int number_of_displayed_views_in_sequence(int number_of_views);
int mirror_sequence_view(int number_of_views, int view);

void update_colors_effected_by_hue_change(struct color_table *colors, struct hue_change *hue_change);
void colorize_color_table(struct color_table *original, struct color_table *colorized,
	struct hue_change *hue_changes, struct rgb_color *replacement_colors, int hue_change_count);

#ifdef __cplusplus
}
#endif

#endif // __SHAPE_FILE_H__
